"""Registration code of Gym environments in this package."""
import math
import gym
from .smb_env import SuperMarioBrosEnv


def _register_mario_env(id: str, **kwargs: dict) -> None:
    """
    Register a Super Mario Bros. (1/2) environment with OpenAI Gym.

    Args:
        id: the id for the env to register
        kwargs: the keyword arguments for the SuperMarioBrosEnv initializer

    Returns:
        None

    """
    gym.envs.registration.register(
        id=id,
        entry_point='gym_super_mario_bros:SuperMarioBrosEnv',
        max_episode_steps=9999999,
        reward_threshold=32000,
        kwargs={ 'max_episode_steps': math.inf, **kwargs},
        nondeterministic=True,
    )


# Super Mario Bros. with standard frame skip
_register_mario_env('SuperMarioBros-v0', frame_skip=4)
_register_mario_env('SuperMarioBros-v1', frame_skip=4, rom_mode='downsample')
_register_mario_env('SuperMarioBros-v2', frame_skip=4, rom_mode='pixel')
_register_mario_env('SuperMarioBros-v3', frame_skip=4, rom_mode='rectangle')


# Super Mario Bros. with no frame skip
_register_mario_env('SuperMarioBrosNoFrameskip-v0', frame_skip=1)
_register_mario_env('SuperMarioBrosNoFrameskip-v1', frame_skip=1, rom_mode='downsample')
_register_mario_env('SuperMarioBrosNoFrameskip-v2', frame_skip=1, rom_mode='pixel')
_register_mario_env('SuperMarioBrosNoFrameskip-v3', frame_skip=1, rom_mode='rectangle')


# Super Mario Bros. 2 (Lost Levels) with standard frame skip
_register_mario_env('SuperMarioBros2-v0', lost_levels=True, frame_skip=4)
_register_mario_env('SuperMarioBros2-v1', lost_levels=True, frame_skip=4, rom_mode='downsample')


# Super Mario Bros. 2 (Lost Levels) with no frame skip
_register_mario_env('SuperMarioBros2NoFrameskip-v0', lost_levels=True, frame_skip=1)
_register_mario_env('SuperMarioBros2NoFrameskip-v1', lost_levels=True, frame_skip=1, rom_mode='downsample')


def make(environment: str) -> gym.Env:
    """Make the environment and return it. same as `gym.make`."""
    return gym.make(environment)


# define the outward facing API of this module (none, gym provides the API)
__all__ = [make.__name__]



"""An environment to skip k frames and return a max between the last two."""
import gym
import numpy as np


class MaxFrameskipEnv(gym.Wrapper):
    """An environment to skip k frames and return a max between the last two."""

    def __init__(self, env, skip: int=4) -> None:
        """
        Initialize a new max frame skip env around an existing environment.

        Args:
            env: the environment to wrap around
            skip: the number of frames to skip (i.e. hold an action for)

        Returns:
            None

        """
        gym.Wrapper.__init__(self, env)
        # most recent raw observations (for max pooling across time steps)
        self._obs_buffer = np.zeros((2, *env.observation_space.shape), dtype=np.uint8)
        self._skip = skip

    def step(self, action):
        """Repeat action, sum reward, and max over last observations."""
        # the total reward from `skip` frames having `action` held on them
        total_reward = 0.0
        done = None
        # perform the action `skip` times
        for i in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            total_reward += reward
            # assign the buffer with the last two frames
            if i == self._skip - 2:
                self._obs_buffer[0] = obs
            if i == self._skip - 1:
                self._obs_buffer[1] = obs
            # break the loop if the game terminated
            if done:
                break
        # Note that the observation on the done=True frame doesn't matter
        # (because the next state isn't evaluated when done is true)
        max_frame = self._obs_buffer.max(axis=0)

        return max_frame, total_reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


# explicitly define the outward facing API of this module
__all__ = [MaxFrameskipEnv.__name__]



"""An environment wrapper to stack observations into a tensor."""
from collections import deque
import numpy as np
import gym


class FrameStackEnv(gym.Wrapper):
    """An environment wrapper to stack observations into a tensor."""

    def __init__(self, env, k):
        """Stack k last frames.
        Returns lazy array, which is much more memory efficient.
        See Also
        --------
        baselines.common.atari_wrappers.LazyFrames
        """
        gym.Wrapper.__init__(self, env)
        self.k = k
        self.frames = deque([], maxlen=k)
        shp = env.observation_space.shape
        self.observation_space = gym.spaces.Box(
            low=0,
            high=255,
            shape=(shp[0], shp[1], shp[2] * k),
            dtype=np.uint8
        )

    def reset(self):
        ob = self.env.reset()
        for _ in range(self.k):
            self.frames.append(ob)
        return self._get_ob()

    def step(self, action):
        ob, reward, done, info = self.env.step(action)
        self.frames.append(ob)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.k
        return LazyFrames(list(self.frames))


class LazyFrames(object):
    """A memory efficient buffer for frame tensors."""

    def __init__(self, frames):
        """
        This object ensures that common frames between the observations are
        only stored once. It exists purely to optimize memory usage which can
        be huge for DQN's 1M frames replay buffers. This object should only be
        converted to numpy array before being passed to the model. You'd not
        believe how complex the previous solution was.
        """
        self._frames = frames
        self._out = None

    def _force(self):
        if self._out is None:
            self._out = np.concatenate(self._frames, axis=2)
            self._frames = None
        return self._out

    def __array__(self, dtype=None):
        out = self._force()
        if dtype is not None:
            out = out.astype(dtype)
        return out

    def __len__(self):
        return len(self._force())

    def __getitem__(self, i):
        return self._force()[i]


# explicitly define the outward facing API of this module
__all__ = [FrameStackEnv.__name__]



"""A gym wrapper for caching rewards."""
import gym


class RewardCacheEnv(gym.Wrapper):
    """a wrapper that caches rewards of episodes."""

    def __init__(self, env) -> None:
        """
        Initialize a reward caching environment.

        Args:
            env: the environment to wrap

        Returns:
            None

        """
        gym.Wrapper.__init__(self, env)
        self._score = 0
        self.env.unwrapped.episode_rewards = []

    def step(self, action):
        state, reward, done, info = self.env.step(action)
        self._score += reward
        if done:
            self.env.unwrapped.episode_rewards.append(self._score)
            self._score = 0
        return state, reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


# explicitly specify the external API of this module
__all__ = [RewardCacheEnv.__name__]



"""Methods for playing the game randomly, or as a human."""
from tqdm import tqdm
import gym
from ._human_play import play


def play_human(env: gym.Env) -> None:
    """
    Play the environment using keyboard as a human.

    Args:
        env: the initialized gym environment to play

    Returns:
        None

    """
    # play the game and catch a potential keyboard interrupt
    try:
        play(env)
    except KeyboardInterrupt:
        pass
    # reset and close the environment
    env.reset()
    env.close()


def play_random(env: gym.Env) -> None:
    """
    Play the environment making uniformly random decisions.

    Args:
        env: the initialized gym environment to play

    Returns:
        None
    """
    try:
        done = True
        progress = tqdm(range(500))
        for step in progress:
            if done:
                state = env.reset()
            action = env.action_space.sample()
            state, reward, done, info = env.step(action)
            progress.set_postfix(reward=reward)
    except KeyboardInterrupt:
        pass
    # reset and close the environment
    env.reset()
    env.close()


# explicitly define the outward facing API of this module
__all__ = [
    play_human.__name__,
    play_random.__name__,
]



"""A gym wrapper for penalizing deaths."""
import gym


class PenalizeDeathEnv(gym.Wrapper):
    """a wrapper that penalizes deaths, without terminating episodes."""

    def __init__(self, env, penalty: int=-1) -> None:
        """
        Initialize a new death penalizing environment wrapper.

        Args:
            env: the environment to wrap
            penalty: the penalty for losing a life

        Returns:
            None

        """
        gym.Wrapper.__init__(self, env)
        self.penalty = penalty

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        reward = self.penalty if done else reward

        return obs, reward, done, info

    def reset(self):
        return self.env.reset()


# explicitly specify the external API of this module
__all__ = [PenalizeDeathEnv.__name__]



"""An environment wrapper for down-sampling frames from RGB to smaller B&W."""
import gym
import cv2
import numpy as np


class DownsampleEnv(gym.ObservationWrapper):
    """An environment that down-samples frames."""

    def __init__(self, env, image_size: tuple):
        """
        Create a new down-sampler.

        Args:
            image_size: the size to output frames as

        Returns:
            None

        """
        super().__init__(env)
        self.image_size = image_size
        # set up a new observation space
        self.observation_space = gym.spaces.Box(
            low=0,
            high=255,
            shape=(image_size[1], image_size[0], 1),
            dtype=np.uint8
        )

    def observation(self, frame):
        # convert the frame from RGB to gray scale
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        # resize the frame to the expected shape. use bilinear interpolation
        frame = cv2.resize(frame, self.image_size)

        return frame[:, :, np.newaxis]


# explicitly define the outward facing API of this module
__all__ = [DownsampleEnv.__name__]



"""Errors for the project."""


class DependencyNotFoundError(OSError):
    """An error signaling that a necessary dependency was not found."""
    pass


__all__ = [
    DependencyNotFoundError.__name__
]



"""Registration code of Gym environments in this package."""
from .smb_env import SuperMarioBrosEnv
from ._registration import make
from .wrappers import wrap


# define the outward facing API of this module (none, gym provides the API)
__all__ = [
    SuperMarioBrosEnv.__name__,
    make.__name__,
    wrap.__name__,
]



"""Super Mario Bros for OpenAI Gym."""
import argparse
from ._registration import make
from ._play import play_human, play_random
from .wrappers import wrap


def create_argparser() -> argparse.ArgumentParser:
    """Create and return an argument parser for this command line interface."""
    parser = argparse.ArgumentParser(description=__doc__)
    # add the argument for the Super Mario Bros environment to run
    parser.add_argument('--env', '-e',
        type=str,
        default='SuperMarioBros-v0',
        choices=[
            'SuperMarioBros-v0',
            'SuperMarioBros-v1',
            'SuperMarioBros-v2',
            'SuperMarioBros-v3',
            'SuperMarioBrosNoFrameskip-v0',
            'SuperMarioBrosNoFrameskip-v1',
            'SuperMarioBrosNoFrameskip-v2',
            'SuperMarioBrosNoFrameskip-v3',
            'SuperMarioBros2-v0',
            'SuperMarioBros2-v1',
            'SuperMarioBros2NoFrameskip-v0',
            'SuperMarioBros2NoFrameskip-v1',
        ],
        help='The name of the environment to play.'
    )
    # add the argument for the mode of execution as either human or random
    parser.add_argument('--mode', '-m',
        type=str,
        default='human',
        choices=['human', 'random'],
        help='The execution mode as either `human` or `random`.'
    )

    return parser


def main() -> None:
    """The main entry point for the command line interface."""
    # parse arguments from the command line (args are validated by argparse)
    args = create_argparser().parse_args()
    # select the method for playing the game
    mode = args.mode
    if mode == 'human':
        play = play_human
    elif mode == 'random':
        play = play_random
    # play the game
    env = make(args.env)
    play(env)


# explicitly define the outward facing API of this module
__all__ = [main.__name__]



"""Wrappers for altering the functionality of the game."""
import gym
from .clip_reward_env import ClipRewardEnv
from .downsample_env import DownsampleEnv
from .frame_stack_env import FrameStackEnv
from .max_frameskip_env import MaxFrameskipEnv
from .monitor import Monitor
from .penalize_death_env import PenalizeDeathEnv
from .reward_cache_env import RewardCacheEnv


def wrap(env: gym.Env,
    image_size: tuple=(100, 100),
    skip_frames: int=None,
    death_penalty: int=-100,
    clip_rewards: bool=False,
    agent_history_length: int=4
) -> gym.Env:
    """
    Wrap an environment with standard wrappers.

    Args:
        env: the environment to wrap
        image_size: the size to down-sample images to
        skip_frames: the number of frames to hold each action for
        death_penatly: the penalty for losing a life in a game
        clip_rewards: whether to clip rewards in {-1, 0, +1}
        agent_history_length: the size of the frame buffer for the agent

    Returns:
        a gym environment configured for this experiment

    """
    # wrap the environment with a reward cacher
    env = RewardCacheEnv(env)
    # apply the frame skip feature if enabled
    if skip_frames is not None:
        env = MaxFrameskipEnv(env, skip=skip_frames)
    # apply a down-sampler for the given game
    env = DownsampleEnv(env, image_size)
    # apply the death penalty feature if enabled
    if death_penalty is not None:
        env = PenalizeDeathEnv(env, penalty=death_penalty)
    # clip the rewards in {-1, 0, +1} if the feature is enabled
    if clip_rewards:
        env = ClipRewardEnv(env)
    # apply the back history of frames if the feature is enabled
    if agent_history_length is not None:
        env = FrameStackEnv(env, agent_history_length)

    return env


# explicitly define the outward facing API of this package
__all__ = [
    ClipRewardEnv.__name__,
    DownsampleEnv.__name__,
    FrameStackEnv.__name__,
    MaxFrameskipEnv.__name__,
    Monitor.__name__,
    PenalizeDeathEnv.__name__,
    RewardCacheEnv.__name__,
    wrap.__name__,
]



"""An environment wrapper to clip rewards."""
import gym
import numpy as np


class ClipRewardEnv(gym.RewardWrapper):
    """An environment that clips rewards in {-1, 0, 1}."""

    def __init__(self, env):
        """Initialize a new reward clipping environment."""
        gym.RewardWrapper.__init__(self, env)

    def reward(self, reward):
        """Bin reward to {-1, 0, +1} using its sign."""
        return np.sign(reward)


# explicitly specify the external API of this module
__all__ = [ClipRewardEnv.__name__]



"""A monitor for capturing experiences from game-play."""
import os
from datetime import datetime
import gym
import cv2


def write_video(frames: list, file_path: str, fps: int) -> None:
    """
    Write a video to disk.

    Args:
        frames: a list of numpy frames of the same shape representing frames
        file_path: the path of the file to create

    Returns:
        None

    """
    # cant write an empty video
    if len(frames) == 0:
        return
    # setup a new video writer with cv2
    height, width, channels = frames[0].shape
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    video = cv2.VideoWriter(file_path, fourcc, fps, (width, height))
    # write the frames to the video file
    for frame in frames:
        # cv2 wants BGR for some reason?
        video.write(frame[:, :, ::-1])
    # cleanup
    video.release()
    cv2.destroyAllWindows()


class Monitor(gym.Wrapper):
    """A monitor for capturing experiences from game-play."""

    def __init__(self, env, out_dir: str) -> None:
        """
        Initialize a new monitor around an existing environment.

        Args:
            env: the environment to capture game-play from
            out_dir: the directory to write monitoring results to

        Returns:
            None

        """
        super().__init__(env)
        self.out_dir = out_dir
        self.episode = -1
        # setup the timestamped output directory
        now = datetime.today().strftime('%Y-%m-%d_%H-%M-%S')
        self.stamped_out_dir = '{}/{}'.format(self.out_dir, now)
        if not os.path.exists(self.stamped_out_dir):
            os.makedirs(self.stamped_out_dir)
        self.frames = None

    @property
    def current_video_path(self) -> str:
        """Return the path to the current video file."""
        return '{}/{}.mp4'.format(self.stamped_out_dir, self.episode)

    def reset(self) -> any:
        """Reset the environment and return the initial observation."""
        self.episode += 1
        # self.frames is None before the first episode
        if self.frames is not None:
            write_video(
                self.frames,
                self.current_video_path,
                self.env.unwrapped.metadata['video.frames_per_second']
            )
        observation = self.env.reset()
        self.frames = [observation]

        return observation

    def step(self, action: any) -> tuple:
        """
        Take a step using the given action.

        Args:
            action: the action to perform

        Returns:
            a tuple of:
            -   the observation as a result of the action
            -   the reward achieved by taking the action
            -   a flag denoting whether the episode has ended
            -   a dictionary of additional information

        """
        observation, reward, done, info = self.env.step(action)
        self.frames.append(observation)

        return observation, reward, done, info


__all__ = [Monitor.__name__]



import os
from gym import spaces
from .nes_env import NESEnv


class SuperMarioBrosEnv(NESEnv):
    """An environment for playing Super Mario Bros with OpenAI Gym."""

    def __init__(self,
        rom_mode: str=None,
        lost_levels: bool=False,
        **kwargs
    ) -> None:
        """
        Initialize a new Super Mario Bros environment.

        Args:
            rom_mode: the ROM mode to use when loading ROMs from disk. valid
                options are:
                - 'standard': the standard ROM with no modifications
                - 'downsample': down-sampled ROM with static artifacts removed
                - 'pixel': a simpler pixelated version of graphics
                - 'rectangle': an even simpler rectangular version of graphics

        Returns:
            None

        """
        super().__init__(**kwargs)
        # load the package directory of this class
        package_directory = os.path.dirname(os.path.abspath(__file__))
        # setup the path to the Lua script
        lua_name = 'lua/super-mario-bros.lua'
        self.lua_interface_path = os.path.join(package_directory, lua_name)
        # setup the path to the game ROM
        if lost_levels:
            if rom_mode is None:
                rom_name = 'roms/super-mario-bros-2.nes'
            elif rom_mode == 'pixel':
                raise ValueError('pixel_rom not supported for Lost Levels')
            elif rom_mode == 'rectangle':
                raise ValueError('rectangle_rom not supported for Lost Levels')
            elif rom_mode == 'downsample':
                rom_name = 'roms/super-mario-bros-2-downsampled.nes'
            else:
                raise ValueError('invalid rom_mode ({})'.format(rom_mode))
        else:
            if rom_mode is None:
                rom_name = 'roms/super-mario-bros.nes'
            elif rom_mode == 'pixel':
                rom_name = 'roms/super-mario-bros-pixel.nes'
            elif rom_mode == 'rectangle':
                rom_name = 'roms/super-mario-bros-rect.nes'
            elif rom_mode == 'downsample':
                rom_name = 'roms/super-mario-bros-downsampled.nes'
            else:
                raise ValueError('invalid rom_mode ({})'.format(rom_mode))
        # convert the path to an absolute path
        self.rom_file_path = os.path.join(package_directory, rom_name)
        # setup the discrete action space for the agent
        self.actions = [
            '',    # NOP
            'U',   # Up
            'D',   # Down
            'L',   # Left
            'R',   # Right
            'LA',  # Left + A
            'LB',  # Left + B
            'LAB', # Left + A + B
            'RA',  # Right + A
            'RB',  # Right + B
            'RAB', # Right + A + B
            'A',   # A
            'B',   # B
            'AB'   # A + B
        ]
        self.action_space = spaces.Discrete(len(self.actions))

    def get_keys_to_action(self) -> dict:
        """Return the dictionary of keyboard keys to actions."""
        # Mapping of buttons on the NES joy-pad to keyboard keys
        up =    ord('w')
        down =  ord('s')
        left =  ord('a')
        right = ord('d')
        A =     ord('o')
        B =     ord('p')
        # A mapping of pressed key combinations to discrete actions in action space
        keys_to_action = {
            (): 0,
            (up, ): 1,
            (down, ): 2,
            (left, ): 3,
            (right, ): 4,
            tuple(sorted((left, A, ))): 5,
            tuple(sorted((left, B, ))): 6,
            tuple(sorted((left, A, B, ))): 7,
            tuple(sorted((right, A, ))): 8,
            tuple(sorted((right, B, ))): 9,
            tuple(sorted((right, A, B, ))): 10,
            (A, ): 11,
            (B, ): 12,
            tuple(sorted((A, B))): 13
        }

        return keys_to_action


__all__ = [SuperMarioBrosEnv.__name__]



"""A method to play gym environments using human IO inputs."""
import os
import gym
import pygame
from typing import Callable


def display_arr(screen, arr, video_size, transpose):
    arr_min, arr_max = arr.min(), arr.max()
    arr = 255.0 * (arr - arr_min) / (arr_max - arr_min)
    pyg_img = pygame.surfarray.make_surface(arr.swapaxes(0, 1) if transpose else arr)
    pyg_img = pygame.transform.scale(pyg_img, video_size)
    screen.blit(pyg_img, (0,0))


def play(env: gym.Env,
    transpose: bool=True,
    fps: int=30,
    callback: Callable=None,
    nop_: any=0,
) -> None:
    """Play the game using the keyboard as a human.

    Args:
        env: gym.Env
            Environment to use for playing.
        transpose: bool
            If True the output of observation is transposed.
            Defaults to true.
        fps: int
            Maximum number of steps of the environment to execute every second.
            Defaults to 30.
        callback: lambda or None
            Callback if a callback is provided it will be executed after
            every step. It takes the following input:
                obs_t: observation before performing action
                obs_tp1: observation after performing action
                action: action that was executed
                rew: reward that was received
                done: whether the environemnt is done or not
                info: debug info
        nop_: the object to use as a null op action for the environment

    Returns:
        None

    """
    # type check the observation space
    obs_s = env.observation_space
    assert type(obs_s) == gym.spaces.box.Box
    assert len(obs_s.shape) == 2 or (len(obs_s.shape) == 3 and obs_s.shape[2] in [1,3])
    # get the mapping of keyboard keys to actions in the environment
    if hasattr(env, 'get_keys_to_action'):
        keys_to_action = env.get_keys_to_action()
    elif hasattr(env.unwrapped, 'get_keys_to_action'):
        keys_to_action = env.unwrapped.get_keys_to_action()
    else:
        raise ValueError('env has no get_keys_to_action method')
    relevant_keys = set(sum(map(list, keys_to_action.keys()),[]))
    # transpose the video is specified
    if transpose:
        video_size = env.observation_space.shape[1], env.observation_space.shape[0]
    else:
        video_size = env.observation_space.shape[0], env.observation_space.shape[1]

    pressed_keys = []
    running = True
    env_done = True
    # setup the screen using pygame
    screen = pygame.display.set_mode(video_size)
    pygame.display.set_caption(env.spec.id)
    # disable the SDL video driver so FCEUX wont open a window
    os.environ['SDL_VIDEODRIVER'] = 'dummy'
    clock = pygame.time.Clock()
    # start the main game loop
    while running:
        if env_done:
            env_done = False
            obs = env.reset()
        else:
            action = keys_to_action.get(tuple(sorted(pressed_keys)), nop_)
            prev_obs = obs
            obs, rew, env_done, info = env.step(action)
            if callback is not None:
                callback(prev_obs, obs, action, rew, env_done, info)
        if obs is not None:
            if len(obs.shape) == 2:
                obs = obs[:, :, None]
            if obs.shape[2] == 1:
                obs = obs.repeat(3, axis=2)
            display_arr(screen, obs, transpose=transpose, video_size=video_size)

        # process pygame events
        for event in pygame.event.get():
            # test events, set key states
            if event.type == pygame.KEYDOWN:
                if event.key in relevant_keys:
                    pressed_keys.append(event.key)
                elif event.key == 27:
                    running = False
            elif event.type == pygame.KEYUP:
                if event.key in relevant_keys:
                    pressed_keys.remove(event.key)
            elif event.type == pygame.QUIT:
                running = False

        pygame.display.flip()
        clock.tick(fps)
    pygame.quit()


# explicitly define the outward facing API of the module
__all__ = [play.__name__]
